/* $Id: 1$ */
/** @FindAddress.c
  *
  * Find module in memory and return address to given export
  *
  * Author: Pavel Shtemenko <pasha@fortuna.odessa.ua>
  *
  *
  */
/******************************************************************************
 *
 * 1. Copyright Notice
 *
 * Some or all of this work - Copyright (c) 1999 - 2008, Intel Corp.
 * All rights reserved.
 *
 * 2. License
 *
 * 2.1. This is your license from Intel Corp. under its intellectual property
 * rights.  You may have additional license terms from the party that provided
 * you this software, covering your right to use that party's intellectual
 * property rights.
 *
 * 2.2. Intel grants, free of charge, to any person ("Licensee") obtaining a
 * copy of the source code appearing in this file ("Covered Code") an
 * irrevocable, perpetual, worldwide license under Intel's copyrights in the
 * base code distributed originally by Intel ("Original Intel Code") to copy,
 * make derivatives, distribute, use and display any portion of the Covered
 * Code in any form, with the right to sublicense such rights; and
 *
 * 2.3. Intel grants Licensee a non-exclusive and non-transferable patent
 * license (with the right to sublicense), under only those claims of Intel
 * patents that are infringed by the Original Intel Code, to make, use, sell,
 * offer to sell, and import the Covered Code and derivative works thereof
 * solely to the minimum extent necessary to exercise the above copyright
 * license, and in no event shall the patent license extend to any additions
 * to or modifications of the Original Intel Code.  No other license or right
 * is granted directly or by implication, estoppel or otherwise;
 *
 * The above copyright and patent license is granted only if the following
 * conditions are met:
 *
 * 3. Conditions
 *
 * 3.1. Redistribution of Source with Rights to Further Distribute Source.
 * Redistribution of source code of any substantial portion of the Covered
 * Code or modification with rights to further distribute source must include
 * the above Copyright Notice, the above License, this list of Conditions,
 * and the following Disclaimer and Export Compliance provision.  In addition,
 * Licensee must cause all Covered Code to which Licensee contributes to
 * contain a file documenting the changes Licensee made to create that Covered
 * Code and the date of any change.  Licensee must include in that file the
 * documentation of any changes made by any predecessor Licensee.  Licensee
 * must include a prominent statement that the modification is derived,
 * directly or indirectly, from Original Intel Code.
 *
 * 3.2. Redistribution of Source with no Rights to Further Distribute Source.
 * Redistribution of source code of any substantial portion of the Covered
 * Code or modification without rights to further distribute source must
 * include the following Disclaimer and Export Compliance provision in the
 * documentation and/or other materials provided with distribution.  In
 * addition, Licensee may not authorize further sublicense of source of any
 * portion of the Covered Code, and must include terms to the effect that the
 * license from Licensee to its licensee is limited to the intellectual
 * property embodied in the software Licensee provides to its licensee, and
 * not to intellectual property embodied in modifications its licensee may
 * make.
 *
 * 3.3. Redistribution of Executable. Redistribution in executable form of any
 * substantial portion of the Covered Code or modification must reproduce the
 * above Copyright Notice, and the following Disclaimer and Export Compliance
 * provision in the documentation and/or other materials provided with the
 * distribution.
 *
 * 3.4. Intel retains all right, title, and interest in and to the Original
 * Intel Code.
 *
 * 3.5. Neither the name Intel nor any other trademark owned or controlled by
 * Intel shall be used in advertising or otherwise to promote the sale, use or
 * other dealings in products derived from or relating to the Covered Code
 * without prior written authorization from Intel.
 *
 * 4. Disclaimer and Export Compliance
 *
 * 4.1. INTEL MAKES NO WARRANTY OF ANY KIND REGARDING ANY SOFTWARE PROVIDED
 * HERE.  ANY SOFTWARE ORIGINATING FROM INTEL OR DERIVED FROM INTEL SOFTWARE
 * IS PROVIDED "AS IS," AND INTEL WILL NOT PROVIDE ANY SUPPORT,  ASSISTANCE,
 * INSTALLATION, TRAINING OR OTHER SERVICES.  INTEL WILL NOT PROVIDE ANY
 * UPDATES, ENHANCEMENTS OR EXTENSIONS.  INTEL SPECIFICALLY DISCLAIMS ANY
 * IMPLIED WARRANTIES OF MERCHANTABILITY, NONINFRINGEMENT AND FITNESS FOR A
 * PARTICULAR PURPOSE.
 *
 * 4.2. IN NO EVENT SHALL INTEL HAVE ANY LIABILITY TO LICENSEE, ITS LICENSEES
 * OR ANY OTHER THIRD PARTY, FOR ANY LOST PROFITS, LOST DATA, LOSS OF USE OR
 * COSTS OF PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES, OR FOR ANY INDIRECT,
 * SPECIAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THIS AGREEMENT, UNDER ANY
 * CAUSE OF ACTION OR THEORY OF LIABILITY, AND IRRESPECTIVE OF WHETHER INTEL
 * HAS ADVANCE NOTICE OF THE POSSIBILITY OF SUCH DAMAGES.  THESE LIMITATIONS
 * SHALL APPLY NOTWITHSTANDING THE FAILURE OF THE ESSENTIAL PURPOSE OF ANY
 * LIMITED REMEDY.
 *
 * 4.3. Licensee shall not export, either directly or indirectly, any of this
 * software or system incorporating such software without first obtaining any
 * required license or other approval from the U. S. Department of Commerce or
 * any other agency or department of the United States Government.  In the
 * event Licensee exports any such software from the United States or
 * re-exports any such software from a foreign destination, Licensee shall
 * ensure that the distribution and export/re-export of the software is in
 * compliance with all laws, regulations, orders, or other restrictions of the
 * U.S. Export Administration Regulations. Licensee agrees that neither it nor
 * any of its subsidiaries will export/re-export any technical data, process,
 * software, or service, directly or indirectly, to any country for which the
 * United States government or any agency thereof requires an export license,
 * other governmental approval, or letter of assurance, without first obtaining
 * such license, approval or letter.
 *
 *****************************************************************************/
#define INCL_32
#define INCL_NOBASEAPI
#define INCL_ERROR_H
#include <os2.h>
#include "bseerr.h"
#include "miscx.h"
#include "reqpktx.h"
#include "bsekee.h"
#include "psd.h"
#include "acpi.h"
#include "ipdc.h"
#include "acpiapi.h"
#include "acpifun.h"
#pragma pack(push,1)
#include "sas.h"

typedef struct SMTE
{
   ULONG  smte_mpages;
   ULONG  smte_startobj;
   ULONG  smte_eip;
   ULONG  smte_stackobj;
   ULONG  smte_esp;
   ULONG  smte_pageshift;
   ULONG  smte_fixupsize;
   ULONG  smte_objtab;
   ULONG  smte_objcnt;
   ULONG  smte_objmap;
   ULONG  smte_itermap;
   ULONG  smte_rsrctab;
   ULONG  smte_rsrccnt;
   ULONG  smte_restab;
   ULONG  smte_enttab;
   ULONG  smte_fpagetab;
   ULONG  smte_frectab;
   ULONG  smte_impmod;
   ULONG  smte_impproc;
   ULONG  smte_datapage;
   ULONG  smte_nrestab;
   ULONG  smte_cbnrestab;
   ULONG  smte_autods;
   ULONG  smte_debuginfo;
   ULONG  smte_debuglen;
   ULONG  smte_heapsize;
   ULONG  smte_path;
   ULONG  smte_semcount;
   ULONG  smte_semowner;
   ULONG  smte_pfilecache;
   ULONG  smte_stacksize;
   USHORT smte_alignshift;
   USHORT smte_NEexpver;
   USHORT smte_pathlen;
   USHORT smte_NEexetype;
   USHORT smte_csegpack;
} SMTE, * PSMTE;

typedef struct MTE
{
   USHORT mte_flags2;
   USHORT mte_handle;
   ULONG  mte_swapmte;
   ULONG  mte_link;
   ULONG  mte_flags1;
   ULONG  mte_impmodcnt;
   USHORT mte_sfn;
   USHORT mte_usecnt;
   UCHAR  mte_modname[8];
} MTE, * PMTE;

typedef struct OTE
{
   ULONG ote_size;
   ULONG ote_base;
   ULONG ote_flags;
   ULONG ote_pagemap;
   ULONG ote_mapsize;
   USHORT ote_sel;
   USHORT ote_hob;
} OTE, * POTE;

#pragma pack(pop)
#define OBJEXEC  4

ULONG StrNCmp(UCHAR *, UCHAR *, ULONG);
ULONG StrLen(UCHAR *);

typedef ULONG (PSDCALL *PCALLPSD)(APPCOMMPAR *);
PCALLPSD FindAddress(unsigned char *Name,unsigned char *Export);
/**
* Find address of export 32 bit DLL loaded by kernel
* @Name    DLL name, as it is knowing by kernel
* @Export  Exports name in this dll
* @return  32 bit address of export of find, else NULL 
*/
PCALLPSD 
FindAddress(
    unsigned char *Name,
    unsigned char *Export)
{
    struct SAS *SASsel;
    struct SAS_vm_section *SASvm;
    PMTE   MTE_h;
    PSMTE  SMTE_s;
    POTE   OTE_s;
    ULONG *MTEadr,Addr;
    USHORT *a;
    USHORT Ordinal,TstOrd;
    UCHAR *dll;
    UCHAR ACPIPSD[8]={'A','C','P','I',0,0,0,0};
    UCHAR ii;
    USHORT Sel=0;
    ULONG  Offset=0;

//========= Find PSD =================
    SASsel = (struct SAS *)KernSelToFlat(0x700000);                 // Already?
    SASvm  = (struct SAS_vm_section *)((ULONG)SASsel+(ULONG)SASsel->SAS_vm_data);
    MTEadr = (ULONG *)SASvm->SAS_vm_all_mte;
    MTE_h  = (PMTE)(*MTEadr);

    while (MTE_h)
    {
           if (!StrNCmp(ACPIPSD,MTE_h->mte_modname,8))
           {
               break;    // Found
           }
           MTE_h = (struct MTE *)MTE_h->mte_link;
    }

    if (MTE_h == NULL)
    {
        return NULL;
    }
    MTEadr = (ULONG *)MTE_h;                     // Need address
    SMTE_s = (PSMTE)MTE_h->mte_swapmte;
    dll    = (UCHAR *)SMTE_s->smte_restab;
    if ((dll[(ULONG)dll[0]+1] != 0) || (dll[(ULONG)dll[0]+2] != 0))
    {
         dll += ((ULONG)dll[0] + 3);
    }
    while (0 != dll[0])
    {
           if  (0 == StrNCmp(&dll[1],Export,StrLen(Export)))
           {
                a = (USHORT *)&dll[(dll[0]+1)];
                break;
           }
           dll += ((ULONG)dll[0] + 3);
    }
    if (0 == dll[0])
    {
        return NULL;
    }
    Ordinal = *a;
    a       = (USHORT *)SMTE_s->smte_enttab; // in eax SMTE_s
    for(;;)
    {
        TstOrd = *a;
        if (0 == ((TstOrd) & 0xff))
        {
            break;     //
        }
        if (0 == ((TstOrd) & 0xff00))
        {
            Ordinal -= (TstOrd & 0xff);
            if (Ordinal < ( (TstOrd) & 0xff))
            {
//            dprintf("Not found\n");
                return NULL;
            }
            a++;
            continue;
        }
        if (Ordinal <= ( TstOrd & 0xff))
        {
            ii = ( ( TstOrd & 0x0f00) >> 8);
            if (1 == ii)
            {
                ii = 3;
            }
            else
            {
               if (3 == ii)
               {
                   ii = 5;
               }
               else
               {
                   break;
               }
            }
            Addr   = (ULONG)a;
            Addr  += ((((ULONG)Ordinal - 1) * ((ULONG)ii) )+4 +1); // edx
            MTEadr = (ULONG *)Addr;
            Addr   = *MTEadr;
            if (ii != 5)        // 16-bit offset ?
            {
                Addr &= 0xffff;
            }
            /*
             *  We now need to get the base address and selector of the object.
             *  (TstOrd is now a 32-bit offset into the object)
             */
            a++;
            TstOrd = *a;
            if (0 == TstOrd)        // stroreptr
            {
                Sel=TstOrd;
                Offset=Addr;
//         printf("Found sel: %x offset %x\n",(ULONG)Sel,Offset);
                break;
            }
            if (TstOrd <= SMTE_s->smte_objcnt)   //obj_ok
            {
                OTE_s = (struct OTE *)SMTE_s->smte_objtab;
                OTE_s = &OTE_s[(ULONG)TstOrd-1];               // points to OTE entry.
                if (3 == ii)
                {
                    Sel=OTE_s->ote_sel;
                    Offset=Addr;
                }
                else
                {
                    Addr += OTE_s->ote_base;
                    Offset = Addr;
                }
            }
//         dprintf("Found sel: %x offset %x\n",(ULONG)Sel,Offset);
            break;
        }
//------------------------------------------------------------------------
        ii = ( (TstOrd & 0xff00) >> 8);
        if (1 == ii)
        {
            ii=3;
        }
        else
        {
            if (ii == 2)
            {
                ii=5;
            }
            else
            {
                if (ii == 3)
                {
                    ii=5;
                }
                else
                {
                    if (ii == 4)
                    {
                        ii=7;
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }
        Addr     = (ULONG)a;
        Addr    += ((TstOrd && 0xff) * ii);
        a        = (PUSHORT)Addr;
        a++;
        TstOrd   = *a;
        Ordinal -= (TstOrd && 0xff);
        if (Ordinal < ( (TstOrd) & 0xff))
        {
//         dprintf("Not found\n");
            return NULL;
        }
        a++;
    }

    return (PCALLPSD)Offset;

}
/**
* Compare byte stream
* @first  first byte stream
* @second second byte stream
* @count  how many bytes to compare
* @return compared - 0, not compared - 1
*/
ULONG 
StrNCmp(
    UCHAR *first, 
    UCHAR *second, 
    ULONG count)
{
    ULONG i;

    for (i=0; i < count;i++)
    {
       if (first[i] != second[i])
       {
           return 1;
       }
    }

    return 0;

}
/**
* Length of string
* @first   ACSIIZ string
* @return  Number of symbols in string
*/

ULONG 
StrLen(
    unsigned char *first)
{
    ULONG i;
    for (i=0; i < 255;i++)
    {
         if (0 == first[i])
         {
             return i;
         }
    }

    return 0;

}


