/***************************************************************************
 *      Created 2005  eCo Software                                         *
 *                                                                         *
 *      DISCLAIMER OF WARRANTIES.  The following [enclosed] code is        *
 *      sample code created by eCo Software. This sample code is not part  *
 *      of any standard or eCo Software product and is provided to you     *
 *      solely for the purpose of assisting you in the development of your *
 *      applications.  The code is provided "AS IS", without               *
 *      warranty of any kind. eCo Software shall not be liable for any     *
 *      damages arising out of your use of the sample code, even if they   *
 *      have been advised of the possibility of such damages.              *
 *-------------------------------------------------------------------------*/
#define INCL_DOS
#include <os2.h>
#include <acpi.h>
#include <acpiapi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ipdc.h"

 HFILE hPCI;
void PCISuspendDev (ULONG Bus, ULONG Dev, ULONG Fun);

void 
main(
    ULONG argc,
    char  *argv[])
{
    APIRET          rc;
    ACPI_API_HANDLE Hdl;
    ULONG           i,Num;
    KNOWNDEVICE     dev;
    ACPI_BUFFER     Buffer;
    char            achName[5];
    ACPI_STATUS     Status;
    ULONG Action;

    if (argc < 4)
    {
        printf(" Usage: GetPCIirq bus dev fun\n For example GetPCIirq 0 28 1\n");
        exit(0);
    }

    rc = AcpiStartApi(&Hdl);
    if (rc)
    {
        printf("Error open ACPI API rc=%d - 0x%x\n",rc,rc);
        exit(1);
    }

    dev.PciId.Bus      = atoi(argv[1]);
    dev.PciId.Device   = atoi(argv[2]);
    dev.PciId.Function = atoi(argv[3]);

    rc = AcpiGetPCIDev(&Hdl,&dev);
    if (rc)
    {
        printf(" PCI device with Bus %x Dev %x Fun %x is unknown ACPI\n",
                dev.PciId.Bus, dev.PciId.Device, dev.PciId.Function );
    }
    else
    {
        printf(" PCI device with Bus %x Dev %x Fun %x has\n",
                dev.PciId.Bus, dev.PciId.Device, dev.PciId.Function );
        printf(" IRQ:\n");
        printf("    APIC %d\n",dev.ApicIrq);
        printf("    PIC  %d\n", dev.PicIrq);
        if (dev.Handle)
        {
            memset(achName, 0, 5);
            Buffer.Length  = 5;
            Buffer.Pointer = achName;
            Status = AcpiGetName(dev.Handle, ACPI_SINGLE_NAME, &Buffer);
            if (Status == AE_OK)
                printf(" ACPI name: \"%s\"\n",achName);
            else
                printf("Error get name by handle %x\n",Status);
        }
    }

    rc = AcpiEndApi(&Hdl);

    rc = DosOpen ("OEMHLP$", &hPCI, &Action, 0, 0, FILE_OPEN,
                  OPEN_SHARE_DENYNONE | OPEN_FLAGS_FAIL_ON_ERROR | OPEN_ACCESS_READWRITE,
                  NULL);
    if (!rc)
    {
        PCISuspendDev (dev.PciId.Bus, dev.PciId.Device, dev.PciId.Function);
    }

}

#pragma pack(1)
UINT32  
PCIReadConfigL( 
    char Bus, 
    char Device, 
    char Function, 
    ULONG Reg,
    ULONG *Value) 
{
        struct 
        {
             BYTE SubFunction;
             BYTE Bus, DevFunc;
             BYTE Register;
             BYTE Size;
        }    Parm = {3, 0, 0, 0, 4};
        ULONG ParmLen = sizeof (Parm);
        struct 
        {
              BYTE  rc;
              ULONG Data;
        } Data;

        ULONG DataLen = sizeof (Data);

       if (hPCI) 
       {

           Parm.Bus      = Bus;
           Parm.DevFunc  = (Device << 3) | (Function & 0x07);
           Parm.Register = Reg;

           DosDevIOCtl (hPCI, 0x80, 0x0B, &Parm, ParmLen, &ParmLen, &Data, DataLen, &DataLen);
           if (0 == Data.rc) 
           {
               *Value = Data.Data;
           }
           else
            printf("Data rc = %x\n",Data.rc);
     
           return (Data.rc);
      } 
      printf("No hPCI\n");
      return 0;
}

UINT32
PCIWriteConfigL(
    char Bus, 
    char Device,
    char Function,
    ULONG Reg,
    ULONG Value) 
{
      struct 
      {
            BYTE SubFunction;
            BYTE Bus, DevFunc;
            BYTE Register;
            BYTE Size;
            ULONG Data;
      }     Parm = {4, 0, 0, 0, 4, 0};
      ULONG ParmLen = sizeof (Parm);
      struct 
      {
            BYTE  rc;
      }     Data;
      ULONG DataLen = sizeof (Data);

      if (hPCI) 
      {

          Parm.Bus      = Bus;
          Parm.DevFunc  = (Device << 3) | (Function & 0x07);
          Parm.Register = Reg;
          Parm.Data     = Value;

          DosDevIOCtl (hPCI, 0x80, 0x0B, &Parm, ParmLen, &ParmLen, &Data, DataLen, &DataLen);
          return (Data.rc);
      }
      printf("No hPCI\n");
      return 0;
}

#define PCIREG_CAPABIL          0x34   // d_272028
#define PCI_PM_ID               0x01  // power management capability ID

//PCI Bus Power Management Interface Specification
//power management control/status register
#define PMCSR_PME_STATUS         0x8000
#define PMCSR_PME_EN             0x0100
#define PMCSR_POWER_STATE_MASK   0x0003
#define PMCSR_POWER_STATE_D0     0x0000
#define PMCSR_POWER_STATE_D1     0x0001
#define PMCSR_POWER_STATE_D2     0x0002
#define PMCSR_POWER_STATE_D3     0x0003
//d_272381 end

void PCISuspendDev (ULONG Bus, ULONG Dev, ULONG Fun)
{
    ACPI_PCI_ID  Id;
    ULONG Val;
    ULONG NextCap;
    ULONG NextCap1;

    Id.Bus      = Bus;
    Id.Device   = Dev;
    Id.Function = Fun;

    PCIReadConfigL (Bus,Dev,Fun, 4, &Val);
    if ( Val & 0x100000)
        printf("PCIspnd: read cfg dev is capability (%8.8x)\n",Val);

    // Check PCI APM capabilites
    NextCap = PCIREG_CAPABIL;
    PCIReadConfigL (Bus,Dev,Fun, NextCap, &Val);
    NextCap = Val & 0xff;
    while (NextCap)
    {
          PCIReadConfigL (Bus,Dev,Fun, NextCap, &Val);
          printf("ID:%2.2x  NextPtr:%2.2x PwrCap:%4.4x ",
                  Val & 0xff,
                  (Val >> 8) & 0xff,
                  (Val >> 16) & 0xffff
           );
          NextCap1 = ( (Val >> 8) & 0xff) ;
          PCIReadConfigL (Bus,Dev,Fun, NextCap + 4, &Val);
          printf(" Ctrl+Stat:%4.4x BridgeSupp:%2.2x Data:%2.2x\n",
                  Val & 0xffff,
                  (Val >> 16) & 0xff,
                  (Val >> 23) & 0xff
           );
          NextCap = NextCap1;
    }
#ifdef Drova
    Val |= PMCSR_PME_STATUS;     // PME# is in a de-asserted state
    AcpiOsWritePciConfiguration (&Id, NextCap, Val, 32);
    Val |= PMCSR_PME_EN;         // the device will assert PME# when the PMCSR_PME_STATUS is 1
    AcpiOsWritePciConfiguration (&Id, NextCap, Val, 32);
    Val |= PMCSR_POWER_STATE_D3; // places device into the D3 state
    AcpiOsWritePciConfiguration (&Id, NextCap, Val, 32);
#endif

}
