#define INCL_DOS
#include <OS2.H>
#include <acpi.h>
#include <acpiapi.h>
#include <amlresrc.h>
#include "ipdc.h"
#include <STDIO.H>
#include <STDLIB.H>
#include <CTYPE.H>
#include <MEMORY.H>
#include <STRING.H>

typedef USHORT WORD;
// #undef APIENTRY16
// #define APIENTRY16 _Optlink

// ULONG MapPhysicalToLinear (ULONG PhysicalAddress, ULONG Length);

static UCHAR BusCount;
static UCHAR clearSATAStatus = 0;
static UCHAR VIAspecial = 0;
static UCHAR SISspecial = 0;
static UCHAR verboseScan = FALSE;
static unsigned numAdapters = 0;

/* --------------------------------------------------------------------------*/

#pragma pack(1)

typedef struct _Adapter tAdapter, *pAdapter;
typedef void (*tPostFun)(pAdapter);

typedef struct 
              {
              UINT16 DeviceId;
              UINT16 VendorId;
              char	*Name;
              UINT8  Config;
              tPostFun PostFun;
              tPostFun PostSATA;
              } tAdapterTable, *pAdapterTable;

typedef struct _Adapter 
              {
              UINT8         Bus;
              UINT8         Dev;
              UINT8         Fnc;
              UINT8         Rev;
              UINT16        Vendor;
              UINT16        Device;
              pAdapterTable Adapter;
              ACPI_HANDLE   Handle;
              ACPI_NAME     Name;
              UINT8         PCIIRQ;
              UINT8         PicIRQ;
              UINT8         APicIRQ;
              UINT8         ProgIF;
              UINT8         PCIpin;
              pAdapter      AdapterP;            // For device chain
              pAdapter      AdapterR;            // For router chain
              };

#define MAX_NUM_ADAPTERS 256
static tAdapter Adapter[MAX_NUM_ADAPTERS];

// static int ReadPCIConfigArea (UCHAR Bus, UCHAR Dev, UCHAR Fnc, pPCIConfigArea pC);

int rc;
HFILE hPCI, hIO;
UCHAR Bus, Dev, Fnc;
ULONG BMIFAddr, BMIFLen;
UCHAR BMData[0x10000];

/* --------------------------------------------------------------------------*/

static int PCIReadConfigB (char Bus, char Device, char Function, int Reg,
			   char *Value) {
  struct {
    BYTE SubFunction;
    BYTE Bus, DevFunc;
    BYTE Register;
    BYTE Size;
  } Parm = {3, 0, 0, 0, 1};
  ULONG ParmLen = sizeof (Parm);
  struct {
    BYTE  rc;
    ULONG Data;
  } Data;
  ULONG DataLen = sizeof (Data);

  Parm.Bus	= Bus;
  Parm.DevFunc	= (Device << 3) | (Function & 0x07);
  Parm.Register = Reg;

  DosDevIOCtl (hPCI, 0x80, 0x0B, &Parm, ParmLen, &ParmLen, &Data, DataLen, &DataLen);
  if (0 == Data.rc) {
    *Value = Data.Data;
  }
  return (Data.rc);
}

static int PCIReadConfigW (char Bus, char Device, char Function, int Reg,
			   USHORT *Value) {
  struct {
    BYTE SubFunction;
    BYTE Bus, DevFunc;
    BYTE Register;
    BYTE Size;
  } Parm = {3, 0, 0, 0, 2};
  ULONG ParmLen = sizeof (Parm);
  struct {
    BYTE  rc;
    ULONG Data;
  } Data;
  ULONG DataLen = sizeof (Data);

  Parm.Bus	= Bus;
  Parm.DevFunc	= (Device << 3) | (Function & 0x07);
  Parm.Register = Reg;

  DosDevIOCtl (hPCI, 0x80, 0x0B, &Parm, ParmLen, &ParmLen, &Data, DataLen, &DataLen);
  if (0 == Data.rc) {
    *Value = Data.Data;
  }
  return (Data.rc);
}

static int PCIReadConfigL (char Bus, char Device, char Function, int Reg,
			   ULONG *Value) {
  struct {
    BYTE SubFunction;
    BYTE Bus, DevFunc;
    BYTE Register;
    BYTE Size;
  } Parm = {3, 0, 0, 0, 4};
  ULONG ParmLen = sizeof (Parm);
  struct {
    BYTE  rc;
    ULONG Data;
  } Data;
  ULONG DataLen = sizeof (Data);

  Parm.Bus	= Bus;
  Parm.DevFunc	= (Device << 3) | (Function & 0x07);
  Parm.Register = Reg;

  DosDevIOCtl (hPCI, 0x80, 0x0B, &Parm, ParmLen, &ParmLen, &Data, DataLen, &DataLen);
  if (0 == Data.rc) {
    *Value = Data.Data;
  }
  return (Data.rc);
}

/* --------------------------------------------------------------------------*/

#define MAX_PCI_BUSES	 256
#define MAX_PCI_DEVICES 	  32
#define MAX_PCI_FUNCTIONS	   8

#define PCI_MASS_STORAGE	0x01	/* PCI Base Class code */
#define PCI_IDE_CONTROLLER	0x01	/* PCI Sub Class code  */
#define PCI_RAID_CONTROLLER	0x04	/* PCI Sub Class code  */
#define PCI_SATA_CONTROLLER	0x06	/* PCI Sub Class code  */

#define PCIREG_VENDOR_ID	0x00
#define PCIREG_DEVICE_ID	0x02
#define PCIREG_COMMAND		0x04
#define PCIREG_STATUS		0x06
#define PCIREG_REVISION 	0x08
#define PCIREG_PROGIF		0x09
#define PCIREG_CLASS_CODE	0x0A
#define PCIREG_SUBCLASS 	0x0A
#define PCIREG_CLASS		0x0B
#define PCIREG_CACHE_LINE	0x0C
#define PCIREG_LATENCY		0x0D
#define PCIREG_HEADER_TYPE	0x0E
#define PCIREG_PRIMARY_BUS	0x18
#define PCIREG_SECONDARY_BUS	0x19
#define PCIREG_SUBORDINATE_BUS	0x1A
#define PCIREG_IRQ		0x3C
#define PCIREG_INT_PIN		0x3D

#define PCI_HEADER_TYPE_NORMAL	   0
#define PCI_HEADER_TYPE_BRIDGE	   1
#define PCI_HEADER_TYPE_CARDBUS    2

#define PCI_CLASS_BRIDGE_HOST	 0x0600
#define PCI_CLASS_BRIDGE_ISA	 0x0601
#define PCI_CLASS_BRIDGE_PCI	 0x0604
#define PCI_CLASS_BRIDGE_CARDBUS 0x0607

typedef enum { PICmode = 0, APICmode = 1 } tRouterModes;

typedef struct _IRQRoute 
               {
               ACPI_HANDLE Handle;
               UINT16 Dev;
               UINT16 Fnc;
               UINT8  Pin;
               UINT8  IRQ;
               } tIRQRoute, *pIRQRoute;

typedef struct _Bridge 
              {
              UINT8       Bus;
              UINT8       Dev;
              UINT8       Fnc;
              UINT8       HeaderType;
              UINT16      BridgeType;
              SHORT       PriBus;
              UINT8       SecBus;
              UINT8       SubBus;
              ACPI_HANDLE Handle;
              ACPI_NAME   Name;
              UINT32      numRouterEntries[2];
              pIRQRoute   RouterTable[2];
              } tBridge, *pBridge;

tBridge  Bridges[MAX_PCI_BUSES];
pAdapter AdapterBus[MAX_PCI_BUSES];

///////////////////////////////////////////////////////////////////////////////
UINT8 queryPCIIntPin (UINT8 Bus, UINT8 Dev, UINT8 Fnc) 
{
  UINT8 IntPin;

  PCIReadConfigB (Bus, Dev, Fnc, PCIREG_INT_PIN, &IntPin);
//             printf("PIN 0x%x\n",IntPin);
  switch (IntPin) 
         {
         case 0x01: IntPin = 0; break;
         case 0x02: IntPin = 1; break;
         case 0x03: IntPin = 2; break;
         case 0x04: IntPin = 3; break;
         default  : 
             printf("Error PIN 0x%x\n",IntPin);
             IntPin = 0; 
             break;  /* error in PCI INTx info, assume INTA */
         }
  return (IntPin);
}
pAdapter AddAdapter (UCHAR Bus, UCHAR Dev, UCHAR Fnc) 
{
  pAdapter p,f;

 p = malloc(sizeof(tAdapter));
 if( !p ) return NULL;
 p->Bus      = Bus;
 p->Dev      = Dev;
 p->Fnc      = Fnc;
 p->AdapterP = NULL;
 p->AdapterR = NULL;
 p->Handle   = NULL;

 if( AdapterBus[Bus] )
   {
   f = AdapterBus[Bus];
   while ( f->AdapterP )
         {
         f = f->AdapterP;
         }
   f->AdapterP = p;
   }
 else
   AdapterBus[Bus] = p;

return p;
}

void ScanPCIDevices (void) 
{
  UCHAR  Bus, Dev, Fnc, ucFncMax;
  USHORT Class,Vendor;
  USHORT  Command;
  pBridge pB;
  UCHAR HeaderType;
  UCHAR PriBus, SecBus, SubBus;
  pAdapter pA;

  /* scan all PCI buses beginning with bus 0
   * don't assume that the BIOS has found the correct number of buses
   *
   * keep all bridges to create a spanning tree of the PCI buses
   */

  for (Bus = 0; Bus <= BusCount; Bus++) 
         // on each bus: enumerate all possible PCI devices
      for (Dev = 0; Dev < MAX_PCI_DEVICES; Dev++) 
          {  // enumerate all functions of a given PCI device* attention: some old hardware is lying at us here! 
          ucFncMax = MAX_PCI_FUNCTIONS;  // this is the default
          for (Fnc = 0; Fnc < ucFncMax; Fnc++) 
              {
	       PCIReadConfigW (Bus, Dev, Fnc, PCIREG_CLASS_CODE, &Class);
	       PCIReadConfigB (Bus, Dev, Fnc, PCIREG_HEADER_TYPE, &HeaderType);
	       PCIReadConfigW (Bus, Dev, Fnc, PCIREG_VENDOR_ID, &Vendor);
	/* check for invalid values */
//	       if ( Class      == 0      ) continue;
//	       if ( Class      == 0xFFFF ) continue;
//	       if ( HeaderType == 0xFF   ) continue;
	       if ( Vendor      == 0xFFFF      ) continue;
	       if ( HeaderType == 0xFF   ) HeaderType = 0;

       printf ("Bus:%d Dev:%2d Fnc:%d   Class:%04X   Hdr:%02X\n", Bus, Dev, Fnc, Class, HeaderType);
// this is a single function device enumerate function 0 only  but some old PCI-ISA-bridges have broken PCI config spaces !
	       if ((0 == Fnc) && !(HeaderType & 0x80))
                 { 
	          ucFncMax = 1;
	          if (Class == PCI_CLASS_BRIDGE_ISA) ucFncMax++;
	          }
	       HeaderType &= ~0x80;
              if (HeaderType) 
                 {  // this is a bridge!
	          if (HeaderType <= PCI_HEADER_TYPE_CARDBUS) 
                    {
                    PCIReadConfigB (Bus, Dev, Fnc, PCIREG_PRIMARY_BUS,&PriBus);
                    PCIReadConfigB (Bus, Dev, Fnc, PCIREG_SECONDARY_BUS,&SecBus);
                    PCIReadConfigB (Bus, Dev, Fnc, PCIREG_SUBORDINATE_BUS, &SubBus);
                    if (SecBus == SubBus)
                        printf ("   Bridge %d->%d..%d sec\n", PriBus, SecBus, SubBus);
                    else
                        printf ("   Bridge %d->%d..%d\n", PriBus, SecBus, SubBus);
	    /* check for invalid values */
	             if ((PriBus == 0xFF) || (PriBus != Bus))
                       {
                       printf("Inc Pri\n");
                        continue;
                       }
	             if (SecBus == 0)
                       {
                       printf("Inc Sec\n");
                       continue;
                       }
	             if (SubBus == 0)
                       {
                       printf("Inc Sub\n");
                       continue;
                       }
                    pB             = Bridges + SecBus;
                    pB->Bus        = Bus;
                    pB->Dev        = Dev;
                    pB->Fnc        = Fnc;
                    pB->HeaderType = HeaderType;
                    pB->BridgeType = Class;
                    pB->PriBus     = Bus;
                    pB->SecBus     = SecBus;
                    pB->SubBus     = SubBus;
	    /* adjust number of PCI buses according to found bridge setup */
                    if (BusCount < SubBus) BusCount = SubBus;
	             }
                 }
              else 
                 {
	  /* HOST->PCI bridges have header type 0! */
//                 if ( Class == PCI_CLASS_BRIDGE_HOST ) 
                 if ( ( Class & PCI_CLASS_BRIDGE_HOST ) == PCI_CLASS_BRIDGE_HOST )
                    {

                     PCIReadConfigW (Bus, Dev, Fnc, PCIREG_COMMAND, &Command);
                     if ((Command == 0) || (Command == 0xFFFF)) continue;

	              pB = Bridges + Bus;
                     printf ("   Bridge H->%d\n", Bus);

	              if (!pB->BridgeType) 
                        {
	                 pB->Bus = Bus;
	                 pB->Dev = Dev;
	                 pB->Fnc = Fnc;
	                 pB->HeaderType = HeaderType;
	                 pB->BridgeType = Class;
	                 pB->PriBus = -1;
	                 pB->SecBus = Bus;
	                 pB->SubBus = Bus;
	                 }
	              }
                 else
                    {
                    pA = AddAdapter ( Bus, Dev, Fnc);
                    if ( pA )
                       {
                       PCIReadConfigB (Bus, Dev, Fnc, PCIREG_IRQ,&pA->PCIIRQ);
                       pA->PCIpin = queryPCIIntPin (Bus, Dev, Fnc);
                       }
                    }
	     }
      }
    }
}

/* --------------------------------------------------------------------------*/

#define Val32(x)  ((ULONG)(x))


static int queryIRQforPCIFunction (UCHAR Bus, UCHAR Dev, UCHAR Fnc, tRouterModes Mode) {
  UCHAR IntPin;
  UCHAR IRQ = 0;
  int	PCIBus = Bus;
  pBridge pB;

  IntPin = queryPCIIntPin (Bus, Dev, Fnc);

  while (PCIBus >= 0) {
    pB = Bridges + PCIBus;

    /* if any of the PCI buses on the way up to the host bridge is
     * missing, what to do?
     */

    if (!pB->BridgeType)
      exit;

    if (pB->RouterTable[Mode]) {
      /* this bridge has IRQ router info, look it up */

      pIRQRoute pI;
      for (pI = pB->RouterTable[Mode];
	   pI < (pB->RouterTable[Mode] + pB->numRouterEntries[Mode]);
	   pI++) {
	if ((pI->Dev == Dev) && (pI->Pin == IntPin)) {
	  IRQ = pI->IRQ;
	  PCIBus = -1;
	  break;
	}
      }

    } else {
      /* INT wires are routed to the parent bus
       * update info according to the common rules and retry
       */

      switch (pB->BridgeType) {
	case PCI_CLASS_BRIDGE_HOST :
	  /* what's going on here - we've already reached the root ?
	   * bail out confused ...
	   */
	  break;

	case PCI_CLASS_BRIDGE_CARDBUS :
	  /* let's assume PCI INT routing mode ... */
	  IntPin = queryPCIIntPin (pB->PriBus, pB->Dev, pB->Fnc);
	  break;

	case PCI_CLASS_BRIDGE_PCI :
	  /* assume PCI INT swizzle as per PCI-to-PCI bridge spec. ECN */
	  IntPin = (IntPin + Dev) & 3;
	  break;
      }

      PCIBus = pB->PriBus;
      Dev    = pB->Dev;
      Fnc    = pB->Fnc;
    }
  }

  return (IRQ);
}

int BridgeFromPCIAddress (UCHAR Bus, UCHAR Dev, UCHAR Fnc) 
{
  int i;

  for (i = 0; i <= BusCount; i++) 
      {
      pBridge pB = Bridges + i;
      if ((Bus == pB->Bus) && (Dev == pB->Dev) && (Fnc == pB->Fnc) && (pB->BridgeType != 0))
          return (i);
      }
  return (-1);
}

UCHAR EvaluateMethod (ACPI_HANDLE Handle, char *MethodName, ACPI_BUFFER *Buffer)
{

  Buffer->Length  = 0;
  Buffer->Pointer = NULL;

  if (AE_BUFFER_OVERFLOW == AcpiEvaluateObject (Handle, MethodName, NULL, Buffer)) {
    Buffer->Pointer = malloc (Buffer->Length);
    if (Buffer->Pointer != NULL)
      return (ACPI_SUCCESS (AcpiEvaluateObject (Handle, MethodName, NULL, Buffer)));
  }
  return (FALSE);
}

static UCHAR EvaluateCRSforBus (ACPI_HANDLE Handle, UCHAR *minBus, UCHAR *maxBus) {
  ACPI_BUFFER ResultBuffer;
  UCHAR rc = FALSE;

  if (EvaluateMethod (Handle, METHOD_NAME__CRS, &ResultBuffer)) {
    /* _CRS evaluates to a buffer pointing to the resource byte stream */

    ACPI_OBJECT *pO = (ACPI_OBJECT *)ResultBuffer.Pointer;
    if (pO->Type == ACPI_TYPE_BUFFER) {
      char *Resource = (char *)pO->Buffer.Pointer;
      char *RsrcEnd  = (char *)pO->Buffer.Pointer + pO->Buffer.Length;
      char DescriptorType;

      for (; Resource && (Resource < RsrcEnd);) {
	DescriptorType = *Resource;

	if (DescriptorType == 0x79) { /* end tag */
	  break;

	} else if (DescriptorType == 0x88) { /* WORD address tag */
	  AML_RESOURCE_ADDRESS16 *r = (AML_RESOURCE_ADDRESS16 *)Resource;
	  if (r->ResourceType == 2) { /* BusNumber */
	    rc = (r->Minimum <= 0xFF);
	    if (rc) {
	      *minBus = r->Minimum;
	      rc = ((r->Maximum <= 0xFF) && (r->Minimum <= r->Maximum));
	      if (rc) {
		*maxBus = r->Maximum;
		break;
	      }
	    }
	  }

	} /* we don't care about other address space types because the
	   * 'WordBusNumber' AML macro expands to the 16bit type
	   */

	if (DescriptorType & 0x80)
	  Resource += ((AML_RESOURCE_LARGE_HEADER *)Resource)->ResourceLength
		    + sizeof (AML_RESOURCE_LARGE_HEADER);
	else
	  Resource += (DescriptorType & 0x7)
		    + sizeof (AML_RESOURCE_SMALL_HEADER);
      }
    }
  }
  free (ResultBuffer.Pointer);
  return (rc);
}

/*
 *  Callback for GetIRQDevice, filling IRQ Triggering and Polarity
 * *Context == 1 print
 * *Context == 0 save 1st finding
 */
ACPI_STATUS GetIRQSetting(ACPI_RESOURCE *Resource,
				       void *Context)
{
  UINT32 *IRQ = (UINT32 *) Context;
  UINT32 i;

  switch (Resource->Type)
        {
        case ACPI_RESOURCE_TYPE_IRQ:
             if ( !(*IRQ) )
                  *IRQ = Resource->Data.Irq.Interrupts[0];
             else
                 for (i = 0; i < Resource->Data.Irq.InterruptCount; i++)
                      printf("%2d ",Resource->Data.Irq.Interrupts[i]);
             break;
        case ACPI_RESOURCE_TYPE_EXTENDED_IRQ:
             if ( !(*IRQ) )
                {
                *IRQ = Resource->Data.ExtendedIrq.Interrupts[0];
                }
             else
               for (i=0; i < Resource->Data.ExtendedIrq.InterruptCount; i++)
                   {
                   if ( Resource->Data.ExtendedIrq.Interrupts[i] == 0 ) // Not possible, because IRQ0 HW routed to PIC for Virtual Wire Mode
                      continue;                                       // in APIC and to Timer in PIC.
                   printf("%2d ",Resource->Data.ExtendedIrq.Interrupts[i]);
                   }
            break;
        default:
//            printf("Unknown resource\n");
            break;
        }
return AE_OK;
}
UCHAR EvaluateCRSforIRQ (ACPI_HANDLE Handle) 
{
  ACPI_STATUS Status;
  UINT32 IRQ = 0;

  Status = AcpiWalkResources(Handle, METHOD_NAME__CRS, GetIRQSetting, (void *) &IRQ);     // First current
  if ( Status ) printf("Hdl:%x CRS status %x\n",Handle,Status);

  return ((UCHAR)IRQ);
}
UCHAR EvaluatePRSforIRQ (ACPI_HANDLE Handle) 
{
  ACPI_STATUS Status;
  UINT32 IRQ = 1;

  Status = AcpiWalkResources(Handle, METHOD_NAME__PRS, GetIRQSetting, (void *) &IRQ);     // First current
  if ( Status ) printf("Hdl:%x PRS status %x\n",Handle,Status);
  return (IRQ);
}

void EvaluatePRTElement (pIRQRoute pI, ACPI_OBJECT *pO) 
{
  if ((pO->Type == ACPI_TYPE_PACKAGE) &&
      (pO->Package.Count >= 1)) 
    {

    char  rc;
    UCHAR Dev, Fnc, Pin, IRQ;

    /* this may be a PRT IRQ package element
     * evaluate its fields
     */

    ACPI_OBJECT *F = pO->Package.Elements;

    /* field 0: address on the parent PCI bus */

    rc = (F->Type == ACPI_TYPE_INTEGER);
    if (rc) 
       {
       Dev = Val32 (F->Integer.Value) >> 16;
       rc = (Dev <= MAX_PCI_DEVICES);
       if (rc)
	   pI->Dev = Dev;
       pI->Fnc = Val32 (F->Integer.Value) & 0xffff;
       }
    F++;

    /* field 1: INT pin */

    rc = (F->Type == ACPI_TYPE_INTEGER);
    if (rc) 
       {
       ULONG value = Val32 (F->Integer.Value);
       rc = (value < 0xFF);
       if (rc)
	   pI->Pin = Pin = value;
       }
    F++;

    /* field 2: IRQ source, either zero or a reference to an IRQ producer */

    pI->Handle = 0;
    switch (F->Type) 
           {
           case ACPI_TYPE_INTEGER:
	         rc = (Val32 (F->Integer.Value) == 0);
	         if (rc)
                   {
	            F++;

	  /* field 3: global system interrupt */

	           rc = (F->Type == ACPI_TYPE_INTEGER);
	           if (rc)
	              pI->IRQ = IRQ = Val32 (F->Integer.Value);
	           }
	        break;

          case ACPI_TYPE_ANY:
	        rc = (F->Reference.Handle != NULL);
	        if (rc)
                  {
	           pI->IRQ = IRQ = EvaluateCRSforIRQ (F->Reference.Handle);
                  pI->Handle = F->Reference.Handle;
                  }
	        break;

          default: 
               rc = FALSE;
               break;
          }
     }
}

static UCHAR setRouterMode (tRouterModes Mode) {
  ACPI_STATUS	   Status;
  ACPI_OBJECT	   Obj;
  ACPI_OBJECT_LIST Params;

  Params.Count	 = 1;
  Params.Pointer = &Obj;
  Obj.Type	 = ACPI_TYPE_INTEGER;

  /* Caution: the _PIC method is optional! If it's not present we may
   * safely (as per spec!) assume PIC mode IRQ routing tables to be returned
   */

  Obj.Integer.Value = Mode;

  Status = AcpiEvaluateObject (NULL, "\\_PIC", &Params, NULL);
  if (Mode == PICmode)
    return ((Status == AE_OK) || (Status == AE_NOT_FOUND));
  else
    return (Status == AE_OK);
}

static void EvaluatePRT (pBridge pB, tRouterModes Mode) {
  pIRQRoute RouterTable = NULL;

  pB->RouterTable[Mode]      = NULL;
  pB->numRouterEntries[Mode] = 0;

  /* advise ACPI to return the desired router tables
   * this may fail for modes other than PIC!
   */

  if (setRouterMode (Mode)) {
    ACPI_BUFFER  ResultBuffer = {0, NULL};
    ACPI_OBJECT *Obj;

    /* evaluate _PRT package into a prellocated buffer of sufficient size
     * check for correct type and non-zero result
     * enumerate table elements
     */

    if ((AE_BUFFER_OVERFLOW == AcpiEvaluateObject (pB->Handle, METHOD_NAME__PRT, NULL, &ResultBuffer)) &&
	((Obj = (ACPI_OBJECT *)(ResultBuffer.Pointer = malloc (ResultBuffer.Length))) != NULL) &&
	(ACPI_SUCCESS (AcpiEvaluateObject (pB->Handle, METHOD_NAME__PRT, NULL, &ResultBuffer))) &&
	(Obj->Type == ACPI_TYPE_PACKAGE) &&
	(Obj->Package.Count > 0) &&
	((RouterTable = (pIRQRoute)calloc (Obj->Package.Count, sizeof (RouterTable[0]))) != NULL)) {

      unsigned i;
      ACPI_OBJECT *PRTElement = Obj->Package.Elements;

      /* evaluate each individual PRT table element */

      for (i = 0; i < Obj->Package.Count; i++)
	   EvaluatePRTElement (RouterTable + i, PRTElement + i);

      pB->RouterTable[Mode]	 = RouterTable;
      pB->numRouterEntries[Mode] = i;
    }
    free (Obj);
  }
}

static int hasPRT (ACPI_HANDLE Object) {
  /* check for presence of _PRT member */

  ACPI_STATUS Status;
  ACPI_BUFFER ResultBuffer = {0, NULL};

  Status = AcpiEvaluateObject (Object, METHOD_NAME__PRT, NULL, &ResultBuffer);
  return ((AE_BUFFER_OVERFLOW == Status) || (AE_OK == Status));
}

static int testRootBridge (ACPI_HANDLE Object, ACPI_DEVICE_INFO *Info) 
{

  /* a PCI root bridge is required to
   *  - have a HID or CID of "PNP0A03"
   *  - have a _PRT member
   *  - have a _ADR member (dev.fnc on root PCI bus!)
   *  - _BBN is optional and can generally safely be assumed as 0 because
   *	there is no support for multiple root PCI buses in OS/2
   *  for more on this see http://www.acpi.info/acpi_faq.htm
   */

  static char BridgePnPID[] = "PNP0A03";
  static char BridgePnPIDex[] = "PNP0A08";
  char	  rc = FALSE;
  UCHAR   Bus, Dev, Fnc, minBus, maxBus;
  pBridge pB;

  /* first check HID if present */

  if (Info->Valid & ACPI_VALID_HID)
     {
//     printf(" NHID:[%.9s]", &(Info->HardwareId));
     rc = !strncmp (Info->HardwareId.Value, BridgePnPID, sizeof (BridgePnPID)-1);
     }
// if not try as PCI-E host bridge
  if ( !rc )
    {
//    printf(" HID:[%.9s]", &(Info->HardwareId));
    rc = !strncmp (Info->HardwareId.Value, BridgePnPIDex, sizeof (BridgePnPIDex)-1);
    }

  /* if no success, try CID instead */

  if (!rc && (Info->Valid & ACPI_VALID_CID)) 
     {
     int i;
     for (i = 0; i < Info->CompatibilityId.Count; i++)
         {
         rc = !strncmp (Info->CompatibilityId.Id[i].Value,  BridgePnPID, sizeof (BridgePnPID)-1);
         if ( rc )
             rc = !strncmp (Info->CompatibilityId.Id[i].Value, BridgePnPIDex, sizeof (BridgePnPIDex)-1);
         if (rc) break;
         }
     }

  /* if there's no match in both HID and CID bail out */

  if (!rc) return (-1);

  /* next, check for presence of _PRT member */

  if (!hasPRT (Object)) 
     {
      printf ("test has\'t prt!!!!\n");
      return (-1);
     }

  /* preset some defaults because the following stuff is mostly optional
   * and we are forgiving
   */


  Bus = Dev = Fnc = 0;
  minBus = 0;
  maxBus = (UCHAR)MAX_PCI_BUSES;

  if (Info->Valid & ACPI_VALID_ADR) 
     {
     Dev = Val32 (Info->Address) >> 16;
     Fnc = Val32 (Info->Address) & 0xFFFF;
     }

  /* next, get range of PCI bus numbers accessible through this device */

  if (EvaluateCRSforBus (Object, &minBus, &maxBus))
     Bus = minBus;

  {
    /* next, check for presence and value of the optional _BBN member */

    ACPI_STATUS Status;
    ACPI_OBJECT Value;
    ACPI_BUFFER ResultBuffer;

      ResultBuffer.Length  = sizeof (Value);
//      ResultBuffer.Length  = 1;
      ResultBuffer.Pointer = &Value;

    Status = AcpiEvaluateObject (Object, METHOD_NAME__BBN, NULL, &ResultBuffer);
    if ( Status ) printf ("AcpiEvaluateObject BBN status=%d\n",Status);
    if (ACPI_SUCCESS (Status) && (Value.Type == ACPI_TYPE_INTEGER))
      Bus = Val32 (Value.Integer.Value);
  }

  pB = Bridges + Bus;

  if ( (pB->BridgeType & PCI_CLASS_BRIDGE_HOST) != PCI_CLASS_BRIDGE_HOST) 
     {
     printf("NOT PCI_CLASS_BRIDGE_HOST bus:%d type:%x - %x:%x:%x\n",Bus,pB->BridgeType,pB->Bus, pB->Dev, pB->Fnc);
     return (-1);
     }
  /* in case of multiple host bridges to this bus, sort out the correct one */

  pB->SecBus = minBus;
  pB->SubBus = maxBus;
  pB->Handle = Object;
  pB->Name   = Info->Name;

  EvaluatePRT (pB, PICmode);
  EvaluatePRT (pB, APICmode);

  return (Bus);
}
pAdapter MatchAdapter (UCHAR Bus, UCHAR Dev, UCHAR Fnc) 
{
  pAdapter p,f;

 f = AdapterBus[Bus];
 while ( f )
       {
       if ( f->Dev == Dev && f->Fnc == Fnc ) return f;
       f = f->AdapterP;
       }
return NULL;
}


static void TraverseSystemBus (void) 
{
  ACPI_HANDLE ParentHandle, ChildHandle;
  ACPI_STATUS Status;
  int      Level = 1, MaxDepth = 1000;
  int      PCIBus = -1;
  int      nextPCIBus = -1;
  char Data[1024];
  ACPI_BUFFER Buf;
  UCHAR HeaderType;
  UCHAR Dev, Fnc;
  char examineChilds = TRUE;

  ParentHandle = ACPI_ROOT_OBJECT;
  ChildHandle  = 0;

  /*
   * Traverse the tree of objects until we bubble back up to where we
   * started. When Level is zero, the loop is done because we have
   * bubbled up to (and passed) the original parent handle (StartHandle)
   */

  while (Level > 0) 
        {

    /* Get the next typed object in this scope. Null returned
     * if not found
     */
  ACPI_BUFFER Buffer;
  char achName[5];


    Status = AcpiGetNextObject (ACPI_TYPE_ANY, ParentHandle, ChildHandle, &ChildHandle);
//    if ( Status ) printf ("AcpiGetNextObject Status:%x\n",Status);
    if ( Status == AE_OK) 
       {
       ACPI_OBJECT_TYPE ChildType;
       ACPI_BUFFER Buffer;

       examineChilds = TRUE;
       Buffer.Length = 5;
       Buffer.Pointer = achName;
       Status = AcpiGetName(ChildHandle, ACPI_SINGLE_NAME, &Buffer);
       Status = AcpiGetType (ChildHandle, &ChildType);
//       if ( Status ) printf ("AcpiGetType Status:%x\n",Status);
       if (ChildType == ACPI_TYPE_DEVICE) 
          {
	/* Found a device object */
          ACPI_DEVICE_INFO *Info = (ACPI_DEVICE_INFO *)Data;


//       printf("look %s type %d\n", achName,ChildType);
          Buf.Length  = sizeof (Data);
          Buf.Pointer = Data;
          
	   Status = AcpiGetObjectInfo (ChildHandle, &Buf);
          if ( Status ) printf ("AcpiGetObjectInfo Status:%x\n",Status);
          if ( Status == AE_OK ) 
             {
	      if (PCIBus < 0) 
                {    /* find the PCI root bridges */
	         nextPCIBus = testRootBridge (ChildHandle, Info);
	         } 
             else  // ELSE1
                {
	    /* separate PCI functions and PCI bridges from the rest
	     * get PCI header type
	     */
	         examineChilds = FALSE;

	         if (Info->Valid & ACPI_VALID_ADR) 
                   {
	            Dev = (Val32 (Info->Address) >> 16) & (MAX_PCI_DEVICES - 1);
	            Fnc = Val32 (Info->Address) & (MAX_PCI_FUNCTIONS - 1);
	            PCIReadConfigB (PCIBus, Dev, Fnc, PCIREG_HEADER_TYPE, &HeaderType);
	            HeaderType &= ~0x80;
                   if (HeaderType == 0) 
                      {
                     /* this is a regular PCI function
		        * check if it's in the set of 'interesting' devices
		        * if so, update the device info accordingly
		       */
		        pAdapter pA;
		        pA = MatchAdapter (Bus, Dev, Fnc);
		        if (pA)
                         {
		           pA->Handle = ChildHandle;
		           pA->Name   = Info->Name;
		           }

	             } 
                  else 
                    if (HeaderType <= PCI_HEADER_TYPE_CARDBUS) 
                       {
		         int Bridge = BridgeFromPCIAddress (PCIBus, Dev, Fnc);
		         if (Bridge >= 0)
                          {
		            pBridge pB = Bridges + Bridge;
		  /* this is a viable PCI bridge from the PCI bus scan
		   * update ACPI info
		   * and enumerate childs
		   */

		            pB->Handle = ChildHandle;
		            pB->Name   = Info->Name;
		            if (hasPRT (ChildHandle)) 
                             {
		               EvaluatePRT (pB, PICmode);
		               EvaluatePRT (pB, APICmode);
                             }

		            nextPCIBus	= pB->SecBus;
		            examineChilds = TRUE;
		            }
	                }
	          }  // ELSE1
	      } // if status AcpiGetObjectInfo
	  } // device
      }  // getnextobject

      /* enumerate ACPI child objects if appropriate */

      if ((Level < MaxDepth) && examineChilds) 
         {
//	  if (ACPI_SUCCESS (AcpiGetNextObject (ACPI_TYPE_ANY, ChildHandle, 0, NULL))) 
         Status = AcpiGetNextObject ( ACPI_TYPE_ANY, ChildHandle, 0, NULL);
// if ( Status )
// printf ("AcpiGetNextObject failed status=%x\n",Status);
	  if (ACPI_SUCCESS (Status)) 
            {

	  /* There is at least one child of this object, visit the object */
	      Level++;
	      ParentHandle = ChildHandle;
	      ChildHandle  = 0;
	      PCIBus       = nextPCIBus;
	     }
         Status = AE_OK;
         }
     } 
   else 
     {

      /*
       * No more children in this object (AcpiGetNextObject failed),
       * go back upwards in the namespace tree to the object's parent.
       */
// printf ("AcpiGetNextObject failed\n");
      Level--;
      ChildHandle = ParentHandle;
      AcpiGetParent (ParentHandle, &ParentHandle);
      if (PCIBus >= 0)
	  PCIBus = Bridges[PCIBus].PriBus;
     }
  }
}

void ScanACPITree (void) 
{
  /* we are interested in the PIC mode IRQ routing
   * tell this to the ACPI interpreter
   */

  setRouterMode (PICmode);
  TraverseSystemBus();
}

////////////////////////////////////////////////////////////////////////////////
typedef struct _Router_
               {
               ACPI_HANDLE     Handle;
               struct _Router_ *Next;
               pAdapter        Device;
               } ROUTER, *PROUTER;
PROUTER pRouter = NULL;

void  AddDevToRouter(pAdapter d,ACPI_HANDLE Handle)
{
PROUTER   p;
pAdapter  a;

    p = pRouter;
    if ( p->Handle == 0 )
       {
       p->Handle = Handle;
       p->Device = d;
       return;
       }

while ( ( p->Next ) || (p->Handle == Handle ) )
      {
      if ( p->Handle == Handle )   // Found
         {
         a = p->Device;
         if ( a == NULL )
            {
            p->Device = d;
            return;
            }
         while ( a->AdapterR )
                 a = a->AdapterR;
         a->AdapterR = d;
         return;
         }
      p = p->Next;
      }

   p->Next = malloc(sizeof(ROUTER));
   if ( p->Next == NULL ) return;
   p = p->Next;
   p->Next = 0;
   p->Handle = Handle;
   p->Device = d;

}
/* --------------------------------------------------------------------------*/

int main (int argc, char *argv[]) 
  {
  APIRET rc;
  ULONG Action;
  int i;
  UCHAR isACPIpresent = FALSE;
  UCHAR VersionMajor, VersionMinor;
  pAdapter    p;
  pBridge     pB;
  pAdapter    pA;
  pIRQRoute   irq;
  ACPI_BUFFER Buffer;
  char achName[5];
  ACPI_STATUS Status;
  PROUTER     pR = NULL;


  verboseScan = TRUE;

  rc = DosOpen ("OEMHLP$", &hPCI, &Action, 0, 0, FILE_OPEN,
       OPEN_SHARE_DENYNONE | OPEN_FLAGS_FAIL_ON_ERROR | OPEN_ACCESS_READWRITE,
       NULL);
  if (rc != 0) exit (98);
  rc = DosOpen ("TESTCFG$", &hIO, &Action, 0, 0, FILE_OPEN,
       OPEN_SHARE_DENYNONE | OPEN_FLAGS_FAIL_ON_ERROR | OPEN_ACCESS_READWRITE,
       NULL);

 isACPIpresent = TRUE;

  memset(AdapterBus,0,sizeof(tAdapter)* MAX_PCI_BUSES);

  ScanPCIDevices();
  ScanACPITree();

  pRouter = malloc(sizeof(ROUTER));
  if ( pRouter == NULL ) return 1;
  pRouter->Next = NULL;
  pRouter->Device = NULL;
  pRouter->Handle = 0;

    printf ("Bridges found:\n");
    Bus = 0;
    for (pB = Bridges; pB < (Bridges + MAX_PCI_BUSES); pB++) 
        {
        if (pB->BridgeType)
           {
	    printf ("%2d:%02d.%02d  class:%04X  type: %d  Bus ", pB->Bus, pB->Dev, pB->Fnc, pB->BridgeType, pB->HeaderType);
	    if (pB->PriBus >= 0)
	       printf ("%d->%d", pB->PriBus, pB->SecBus);
	    else
	       printf ("H->%d",  pB->SecBus);
	    if (pB->Handle)
	       printf ("  ACPI: '%.4s' (%08lX) %s", &pB->Name, pB->Handle, pB->numRouterEntries[0] ? "_PRT" : "");
	    printf ("\n");
           Bus = pB->SecBus;
           pA = AdapterBus[Bus];
           while( pA )
                {
                printf("       Dev:%2d (0x%2x) Fnc:%02d INT%c IRQ\'s PCI:%2d ",pA->Dev,pA->Dev,pA->Fnc,pA->PCIpin + 'A',pA->PCIIRQ);
                irq = pB->RouterTable[PICmode];
                for ( i = 0; i < pB->numRouterEntries[PICmode] ; i++)
                    {
                    if ( 
                         ( irq->Dev == pA->Dev    ) &&
                         ( irq->Pin == pA->PCIpin )
                       )
                       {
                       pA ->PicIRQ = irq->IRQ;
                       printf( "PIC:%2d ",irq->IRQ);
                       if ( irq->Handle )
                          {
                          printf("Router:"); 
                          memset(achName, 0, 5);
                          Buffer.Length = 5;
                          Buffer.Pointer = achName;
                          Status = AcpiGetName(irq->Handle, ACPI_SINGLE_NAME, &Buffer);
                          printf("%s ", achName,Status);   fflush(stdout);
                          AddDevToRouter(pA,irq->Handle);
                          }
                       break;
                       }
                    irq++;
                    }
                irq = pB->RouterTable[APICmode];
                for ( i = 0; i < pB->numRouterEntries[APICmode] ; i++)
                    {
                    if ( 
                         ( irq->Dev == pA->Dev    ) &&
                         ( irq->Pin == pA->PCIpin )
                       )
                       {
                       pA ->APicIRQ = irq->IRQ;
                       printf( "APIC:%2d ",irq->IRQ);
                       if ( irq->Handle )
                          {
                          printf("Router:"); 
                          memset(achName, 0, 5);
                          Buffer.Length = 5;
                          Buffer.Pointer = achName;
                          Status = AcpiGetName(irq->Handle, ACPI_SINGLE_NAME, &Buffer);
                          printf("%s ", achName,Status);   fflush(stdout);
                          AddDevToRouter(pA,irq->Handle);
                          }
                       break;
                       }
                    irq++;
                    }
                if ( pA->Handle )
                   {
                   printf("ACPI:"); 
                   memset(achName, 0, 5);
                   Buffer.Length = 5;
                   Buffer.Pointer = achName;
                   Status = AcpiGetName(pA->Handle, ACPI_SINGLE_NAME, &Buffer);
                   printf("%s", achName,Status);   fflush(stdout);
                   }
                printf("\n");
                pA = pA->AdapterP;
                }
          }
       }

  printf ("PCI Bus scan: %d PCI bus%s found.\n", BusCount+1, BusCount ? "es" : "");
  printf ("Sorting by routers:\n");  
  EvaluatePRT (pB, APICmode);
  setRouterMode ( 1 );
  pR = pRouter;
  while ( pR )
        {
        printf("\nRouter: "); 
        memset(achName, 0, 5);
        Buffer.Length = 5;
        Buffer.Pointer = achName;
        Status = AcpiGetName(pR->Handle, ACPI_SINGLE_NAME, &Buffer);
        printf("%s Current IRQ:%2d Possible IRQ: ", achName,EvaluateCRSforIRQ(pR->Handle));
        EvaluatePRSforIRQ(pR->Handle);
        printf("\n");
        pA = pR->Device;
        while ( pA )
              {
              memset(achName, 0, 5); 
              Buffer.Length = 5;
              Buffer.Pointer = achName;
              Status = AcpiGetName(pA->Handle, ACPI_SINGLE_NAME, &Buffer);
              printf("       %4.4s ", achName,Status);
              printf("Bus:%2x Dev:%2d (0x%2x) Fnc:%02d INT%c IRQ\'s PCI:%2d ",pA->Bus,pA->Dev,pA->Dev,pA->Fnc,pA->PCIpin + 'A',pA->PCIIRQ);
              printf("PIC:%2d APIC:%2d\n",pA->PicIRQ, pA->APicIRQ);
              pA = pA->AdapterR;
              }
        pR = pR->Next;
        }
  printf("End Scan\n");
  DosClose (hPCI);
  DosClose (hIO);
  return (0);
}
